//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------
/*
 * processor_preintegrator_force_torque_inertial.h
 *
 *  Created on: Aug 19, 2021
 *      Author: jsola
 */

#ifndef PROCESSOR_PROCESSOR_FORCE_TORQUE_INERTIAL_PREINT_H_
#define PROCESSOR_PROCESSOR_FORCE_TORQUE_INERTIAL_PREINT_H_

#include "bodydynamics/math/force_torque_inertial_delta_tools.h"
#include "bodydynamics/sensor/sensor_force_torque_inertial.h"

#include <core/processor/processor_motion.h>

namespace wolf
{

WOLF_STRUCT_PTR_TYPEDEFS(ParamsProcessorForceTorqueInertialPreint);

struct ParamsProcessorForceTorqueInertialPreint : public ParamsProcessorMotion
{
    ParamsProcessorForceTorqueInertialPreint() = default;
    ParamsProcessorForceTorqueInertialPreint(std::string _unique_name, const ParamsServer& _server)
        : ParamsProcessorMotion(_unique_name, _server)
    {
        n_propellers = _server.getParam<int>(prefix + _unique_name + "/n_propellers");
    }

    std::string print() const override
    {
        return ParamsProcessorMotion::print() + "\n"               //
               + "n_propellers: " + std::to_string(n_propellers);  //
    }

    int n_propellers;
};

WOLF_PTR_TYPEDEFS(ProcessorForceTorqueInertialPreint);
class ProcessorForceTorqueInertialPreint : public ProcessorMotion
{
  public:
    ProcessorForceTorqueInertialPreint(
        ParamsProcessorForceTorqueInertialPreintPtr _params_force_torque_inertial_preint);
    ~ProcessorForceTorqueInertialPreint() override;
    void configure(SensorBasePtr _sensor) override;
    WOLF_PROCESSOR_CREATE(ProcessorForceTorqueInertialPreint, ParamsProcessorForceTorqueInertialPreint);

  protected:
    ParamsProcessorForceTorqueInertialPreintPtr params_force_torque_inertial_preint_;
    SensorForceTorqueInertialPtr                sensor_fti_;

  public:
    /** \brief convert raw CaptureMotion data to the delta-state format.
     *
     * This function accepts raw data and time step dt,
     * and computes the value of the delta-state and its covariance. Note that these values are
     * held by the members delta_ and delta_cov_.
     *
     * @param _data measured motion data
     * @param _data_cov covariance
     * @param _dt time step
     * @param _delta computed delta
     * @param _delta_cov covariance
     * @param _calib current state of the calibrated parameters
     * @param _jacobian_calib Jacobian of the delta wrt calib
     *
     * Rationale:
     *
     * The delta-state format must be compatible for integration using
     * the composition functions doing the math in this class: statePlusDelta() and deltaPlusDelta().
     * See the class documentation for some Eigen::VectorXd suggestions.
     *
     * The data format is generally not the same as the delta format,
     * because it is the format of the raw data provided by the Capture,
     * which is unaware of the needs of this processor.
     *
     * Additionally, sometimes the data format is in the form of a
     * velocity, or a higher derivative, while the delta is in the form of an increment.
     * In such cases, converting from data to delta-state needs integrating
     * the data over the period dt.
     *
     * Two trivial implementations would establish:
     *  - If `_data` is an increment:
     *
     *         delta_     = _data;
     *         delta_cov_ = _data_cov
     *  - If `_data` is a velocity:
     *
     *         delta_     = _data * _dt
     *         delta_cov_ = _data_cov * _dt.
     *
     *  However, other more complicated relations are possible.
     *  In general, we'll have a nonlinear
     *  function relating `delta_` to `_data` and `_dt`, as follows:
     *
     *      delta_ = f ( _data, _dt)
     *
     *  The delta covariance is obtained by classical uncertainty propagation of the data covariance,
     *  that is, through the Jacobians of `f()`,
     *
     *      delta_cov_ = F_data * _data_cov * F_data.transpose
     *
     *  where `F_data = d_f/d_data` is the Jacobian of `f()`.
     */
    void computeCurrentDelta(const Eigen::VectorXd& _data,
                             const Eigen::MatrixXd& _data_cov,
                             const Eigen::VectorXd& _calib,
                             const double           _dt,
                             Eigen::VectorXd&       _delta,
                             Eigen::MatrixXd&       _delta_cov,
                             Eigen::MatrixXd&       _jacobian_calib) const override;

    /** \brief composes a delta-state on top of another delta-state
     * \param _delta1 the first delta-state
     * \param _delta2 the second delta-state
     * \param _dt2 the second delta-state's time delta
     * \param _delta1_plus_delta2 the delta2 composed on top of delta1. It has the format of delta-state.
     *
     * This function implements the composition (+) so that _delta1_plus_delta2 = _delta1 (+) _delta2.
     */
    void deltaPlusDelta(const Eigen::VectorXd& _delta1,
                        const Eigen::VectorXd& _delta2,
                        const double           _dt2,
                        Eigen::VectorXd&       _delta1_plus_delta2) const override;

    /** \brief composes a delta-state on top of another delta-state, and computes the Jacobians
     * \param _delta1 the first delta-state
     * \param _delta2 the second delta-state
     * \param _dt2 the second delta-state's time delta
     * \param _delta1_plus_delta2 the delta2 composed on top of delta1. It has the format of delta-state.
     * \param _jacobian1 the jacobian of the composition w.r.t. _delta1.
     * \param _jacobian2 the jacobian of the composition w.r.t. _delta2.
     *
     * This function implements the composition (+) so that _delta1_plus_delta2 = _delta1 (+) _delta2 and its
     * jacobians.
     */
    void deltaPlusDelta(const Eigen::VectorXd& _delta1,
                        const Eigen::VectorXd& _delta2,
                        const double           _dt2,
                        Eigen::VectorXd&       _delta1_plus_delta2,
                        Eigen::MatrixXd&       _jacobian1,
                        Eigen::MatrixXd&       _jacobian2) const override;

    /** \brief composes a delta-state on top of a state
     * \param _x the initial state
     * \param _delta the delta-state
     * \param _x_plus_delta the updated state. It has the same format as the initial state.
     * \param _dt the time interval spanned by the Delta
     *
     * This function implements the composition (+) so that _x2 = _x1 (+) _delta.
     */
    void statePlusDelta(const VectorComposite& _x,
                        const Eigen::VectorXd& _delta,
                        const double           _dt,
                        VectorComposite&       _x_plus_delta) const override;

    /** \brief Delta zero
     * \return a delta state equivalent to the null motion.
     *
     * Examples (see documentation of the the class for info on Eigen::VectorXd):
     *   - 2d odometry: a 3-vector with all zeros, e.g. VectorXd::Zero(3)
     *   - 3d odometry: delta type is a PQ vector: 7-vector with [0,0,0, 0,0,0,1]
     *   - IMU: PQVBB 10-vector with [0,0,0, 0,0,0,1, 0,0,0] // No biases in the delta !!
     */
    Eigen::VectorXd deltaZero() const override;

    /** \brief correct the preintegrated delta
     * This produces a delta correction according to the appropriate manifold.
     * @param delta_preint : the preintegrated delta to correct
     * @param delta_step : an increment in the tangent space of the manifold
     *
     * We want to do
     *
     *   delta_corr = delta_preint (+) d_step
     *
     * where the operator (+) is the right-plus operator on the delta's manifold.
     *
     * Note: usually in motion pre-integration we have
     *
     *   delta_step = Jac_delta_calib * (calib - calib_pre)
     *
     */
    Eigen::VectorXd correctDelta(const Eigen::VectorXd& _delta_preint,
                                 const Eigen::VectorXd& _delta_step) const override;

  protected:
    /** \brief emplace a CaptureMotion
     * \param _frame_own frame owning the Capture to create
     * \param _sensor Sensor that produced the Capture
     * \param _ts time stamp
     * \param _data a vector of motion data
     * \param _data_cov covariances matrix of the motion data data
     * \param _calib calibration vector
     * \param _calib_preint calibration vector used during pre-integration
     * \param _capture_origin_ptr capture acting as the origin of motions for the MorionBuffer of the created MotionCapture
     */
    CaptureMotionPtr emplaceCapture(const FrameBasePtr&   _frame_own,
                                    const SensorBasePtr&  _sensor,
                                    const TimeStamp&      _ts,
                                    const VectorXd&       _data,
                                    const MatrixXd&       _data_cov,
                                    const VectorXd&       _calib,
                                    const VectorXd&       _calib_preint,
                                    const CaptureBasePtr& _capture_origin_ptr) override;

    /** \brief emplace a feature corresponding to given capture and add the feature to this capture
     * \param _capture_motion: the parent capture
     */
    FeatureBasePtr emplaceFeature(CaptureMotionPtr _capture_own) override;

    /** \brief emplace a factor and link it in the wolf tree
     * \param _feature_motion: the parent feature
     * \param _frame_origin: the frame constrained by this motion factor
     */
    FactorBasePtr emplaceFactor(FeatureBasePtr _feature_motion, CaptureBasePtr _capture_origin) override;

    void setCalibration(const CaptureBasePtr _capture, const VectorXd& _calibration) override;

  public:
    virtual VectorXd getCalibration(const CaptureBaseConstPtr _capture = nullptr) const override;
};

inline Eigen::VectorXd ProcessorForceTorqueInertialPreint::deltaZero() const
{
    return bodydynamics::fti::identity();
}

} /* namespace wolf */

#endif /* PROCESSOR_PROCESSOR_FORCE_TORQUE_INERTIAL_PREINT_H_ */
