//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------
/**
 * \file capture_leg_odom.cpp
 *
 *  Created on: May 12, 2020
 *      \author: mfourmy
 */


#include "core/math/rotations.h"
#include "core/capture/capture_base.h"
#include "bodydynamics/capture/capture_leg_odom.h"


namespace wolf {

// bm: base at time t-dt
// b: base at current time
CaptureLegOdom::CaptureLegOdom(const TimeStamp& _init_ts,
                               SensorBasePtr _sensor_ptr,
                               const Eigen::MatrixXd& _data_kin, // cols: b_p_bl1m, b_p_bl1, b_p_bl2m, b_p_bl2..., rows: xyz
                               Eigen::Matrix6d _data_cov, // for the moment only 6x6 mat, see if more involved computations are necessary (contact confidence...)
                               double dt,
                               CaptureLegOdomType lo_type) : 
                                   CaptureMotion("CaptureLegOdom", _init_ts, _sensor_ptr, 
                                                 Eigen::Vector7d::Zero(), Eigen::Matrix6d::Identity(),  // dummy data and data covariance filled in the constructor
                                                 nullptr, nullptr, nullptr, nullptr)
{
    Vector7d data;
    if (_data_kin.cols() < 2){
        // no feet in contact, flying robot -> dummy data with high cov
        data << 0,0,0, 0,0,0,1;
        _data_cov = pow(1000, 2) * _data_cov;
        return;
    }

    if (lo_type == CaptureLegOdomType::POINT_FEET_RELATIVE_KIN){
        assert(_data_kin.rows() == 3);

        // relative base-feet position + angular velocity
        Eigen::Vector3d i_omg_oi = _data_kin.rightCols<1>();
        Eigen::Quaterniond bm_q_b (v2q(i_omg_oi*dt));
        std::vector<Vector3d> bm_p_bmb_vec;
        for (unsigned int i=0; i < (_data_kin.cols()-1); i+=2){
            Eigen::Vector3d bm_p_bml = _data_kin.col(i);
            Eigen::Vector3d b_p_bl   = _data_kin.col(i+1);
            bm_p_bmb_vec.push_back(bm_p_bml - bm_q_b * b_p_bl);
        }
        // for the moment simple mean, to be improved
        Vector3d bm_p_bmb = Vector3d::Zero();
        for (unsigned int i=0; i < bm_p_bmb_vec.size(); i++){
            bm_p_bmb += bm_p_bmb_vec[i];
        }

        bm_p_bmb /= bm_p_bmb_vec.size();

        data.head<3>() = bm_p_bmb; 
        data.tail<4>() = bm_q_b.coeffs(); 

    }
    else if(lo_type == CaptureLegOdomType::POINT_FEET_DIFFERENTIAL_KIN){
        assert(_data_kin.rows() == 10);
        // TODO: document data matrix organization

        std::vector<Vector3d> b_v_ob_vec;
        Vector3d i_omg_oi = _data_kin.block<3,1>(0, _data_kin.cols()-1);  // retrieve angular vel first
        Eigen::Quaterniond bm_q_b (v2q(i_omg_oi*dt));
        for (unsigned int i=0; i < _data_kin.cols()-1; i++){
            Vector3d b_p_bl = _data_kin.block<3,1>(0,i);
            Quaterniond b_q_l(_data_kin.block<4,1>(3,i));
            Vector3d l_v_bl = _data_kin.block<3,1>(7,i);
            Vector3d b_v_ob = b_p_bl.cross(i_omg_oi) - b_q_l * l_v_bl;
            b_v_ob_vec.push_back(b_v_ob);
        }

        // Simple mean for the moment
        Vector3d b_v_ob_mean = Vector3d::Zero();
        for (unsigned int i=0; i < b_v_ob_vec.size(); i++){
            b_v_ob_mean += b_v_ob_vec[i];
        }
        b_v_ob_mean /= b_v_ob_vec.size();

        data.head<3>() = b_v_ob_mean*dt; 
        data.tail<4>() = bm_q_b.coeffs(); 
    }
    else if(lo_type == CaptureLegOdomType::HUMANOID_FEET_RELATIVE_KIN){
        assert(_data_kin.rows() == 7);
        // relative transformations
        // most simple option: odometry of one of the feet
        Eigen::Vector7d bm_pose_l = _data_kin.col(0);
        Eigen::Vector7d b_pose_l = _data_kin.col(1);
        Eigen::Quaterniond bm_q_l(bm_pose_l.tail<4>());
        Eigen::Quaterniond b_q_l(b_pose_l.tail<4>());
        Eigen::Vector3d bm_p_bml(bm_pose_l.head<3>());
        Eigen::Vector3d b_p_bl(b_pose_l.head<3>());

        Eigen::Quaterniond bm_q_b(bm_q_l * b_q_l.inverse());
        Eigen::Vector3d bm_p_bmb = bm_p_bml - bm_q_b * b_p_bl;

        data.head<3>() = bm_p_bmb; 
        data.tail<4>() = bm_q_b.coeffs(); 
    }
    else {
        std::cout << "Unkown CaptureLegOdomType! ERROR (TODO)" << std::endl;
    }
    setData(data);
    setDataCovariance(_data_cov);
}

CaptureLegOdom::~CaptureLegOdom()
{
    //
}

} //namespace wolf
