//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------

/**
 * \file gtest_processor_point_feet_nomove.cpp
 *
 *  Created on: October 28, 2020
 *      \author: Mederic Fourmy
 */

// debug
#include <iostream>
#include <fstream>

#include <core/utils/utils_gtest.h>
#include <core/utils/logging.h>

// WOLF
#include <core/problem/problem.h>
#include <core/capture/capture_base.h>
// #include <core/ceres_wrapper/solver_ceres.h>
// #include <core/math/rotations.h>
// #include <core/capture/capture_pose.h>
// #include <core/feature/feature_base.h>
// #include <core/factor/factor_block_absolute.h>

// BODYDYNAMICS
#include "bodydynamics/internal/config.h"
#include "bodydynamics/capture/capture_point_feet_nomove.h"
#include "bodydynamics/sensor/sensor_point_feet_nomove.h"
#include "bodydynamics/processor/processor_point_feet_nomove.h"

#include <Eigen/Eigenvalues>

using namespace wolf;
using namespace Eigen;
using namespace std;

const Eigen::Vector3d ZERO3 = Eigen::Vector3d::Zero();
// const Eigen::Vector6d ZERO6 = Eigen::Vector6d::Zero();


class PointFeetCaptures : public testing::Test
{
    public:
        wolf::TimeStamp t_;
        ProblemPtr problem_;
        SensorPointFeetNomovePtr sen_pfnm_;
        ProcessorBasePtr proc_pfnm_base_;
        VectorXd x_origin_;
        // MatrixXd P_origin_;
        FrameBasePtr KF0_;

        CapturePointFeetNomovePtr CPF0_;
        CapturePointFeetNomovePtr CPF1_;
        CapturePointFeetNomovePtr CPF2_;
        CapturePointFeetNomovePtr CPF3_;
        CapturePointFeetNomovePtr CPF4_;
        CapturePointFeetNomovePtr CPF5_;
        CapturePointFeetNomovePtr CPF6_;

    void SetUp() override
    {
        std::string bodydynamics_root_dir = _WOLF_BODYDYNAMICS_ROOT_DIR;

        problem_ = Problem::create("PO", 3);

        // CERES WRAPPER
        // solver_ = std::make_shared<SolverCeres>(problem_);
        // solver_->getSolverOptions().minimizer_type = ceres::TRUST_REGION; //ceres::TRUST_REGION;ceres::LINE_SEARCH
        // solver_->getSolverOptions().max_line_search_step_contraction = 1e-3;
        // solver_->getSolverOptions().max_num_iterations = 1e4;

        //===================================================== INITIALIZATION

        // SENSOR + PROCESSOR POINT FEET NOMOVE
        ParamsSensorPointFeetNomovePtr intr_pfn = std::make_shared<ParamsSensorPointFeetNomove>();
        intr_pfn->std_foot_nomove_ = 0.1;
        VectorXd extr;  // default size for dynamic vector is 0-0 -> what we need
        SensorBasePtr sen_pfnm_base = problem_->installSensor("SensorPointFeetNomove", "SenPfnm", extr, intr_pfn);
        sen_pfnm_ = std::static_pointer_cast<SensorPointFeetNomove>(sen_pfnm_base);
        ParamsProcessorPointFeetNomovePtr params_pfnm = std::make_shared<ParamsProcessorPointFeetNomove>();
        proc_pfnm_base_ = problem_->installProcessor("ProcessorPointFeetNomove", "PPointFeetNomove", sen_pfnm_, params_pfnm);

        // Set origin of the problem
        x_origin_.resize(7);
        x_origin_ << 0,0,0, 0,0,0,1;
        VectorXd std_vec = pow(1e-3, 2)*VectorXd::Ones(6); 
        VectorComposite prior(x_origin_,   "PO", {3,4});
        VectorComposite prior_std(std_vec, "PO", {3,3});
        problem_->setPriorFactor(prior, prior_std, 0.0);
        KF0_ = problem_->getTrajectory()->getLastFrame();

        // Simulate captures with 4 point feet
        std::unordered_map<int, Vector7d> kin_incontact_A({
            {1, Vector7d::Zero()},
            {2, Vector7d::Zero()},
            {3, Vector7d::Zero()},
            {4, Vector7d::Zero()}
        });

        CPF0_ = std::make_shared<CapturePointFeetNomove>(0.0, sen_pfnm_, kin_incontact_A);
        std::cout << CPF0_->getName() << std::endl;
        CPF1_ = std::make_shared<CapturePointFeetNomove>(1.0, sen_pfnm_, kin_incontact_A);
        CPF2_ = std::make_shared<CapturePointFeetNomove>(2.0, sen_pfnm_, kin_incontact_A);

        // contact of the 2nd foot lost
        std::unordered_map<int, Eigen::Vector7d> kin_incontact_B({
            {1, Vector7d::Ones()},
            {3, Vector7d::Ones()},
            {4, Vector7d::Ones()}
        });
        CPF3_ = std::make_shared<CapturePointFeetNomove>(3.0, sen_pfnm_, kin_incontact_B);
        CPF4_ = std::make_shared<CapturePointFeetNomove>(4.0, sen_pfnm_, kin_incontact_B);

        // contact of the 3rd foot lost
        std::unordered_map<int, Eigen::Vector7d> kin_incontact_C({
            {1, 2*Vector7d::Ones()},
            {4, 2*Vector7d::Ones()}
        });
        CPF5_ = std::make_shared<CapturePointFeetNomove>(5.0, sen_pfnm_, kin_incontact_C);
        CPF6_ = std::make_shared<CapturePointFeetNomove>(6.0, sen_pfnm_, kin_incontact_C);

    }

    void TearDown() override{}
};

TEST_F(PointFeetCaptures, process_all_capture_first)
{

    CPF0_->process();
    CPF1_->process();
    CPF2_->process();
    CPF3_->process();
    CPF4_->process();
    CPF5_->process();
    CPF6_->process(); // one last capture to create the factor

    // factor creation due to keyFrameCallback
    FrameBasePtr KF1 = problem_->emplaceFrame(2.0, x_origin_);
    proc_pfnm_base_->keyFrameCallback(KF1);

    FrameBasePtr KF2 = problem_->emplaceFrame(3.0, x_origin_);
    proc_pfnm_base_->keyFrameCallback(KF2);

    problem_->print(4,1,1,1);
}


TEST_F(PointFeetCaptures, process_all_capture_last)
{
    // First, get the key frames from the keyFrameCallback
    FrameBasePtr KF1 = problem_->emplaceFrame(2.0, x_origin_);
    proc_pfnm_base_->keyFrameCallback(KF1);

    FrameBasePtr KF2 = problem_->emplaceFrame(3.0, x_origin_);
    proc_pfnm_base_->keyFrameCallback(KF2);

    // Then process the captures
    CPF0_->process();
    CPF1_->process();
    CPF2_->process();
    CPF3_->process();
    CPF4_->process();
    CPF5_->process();
    CPF6_->process(); // one last capture to create the factor

    problem_->print(4,1,1,1);
}




int main(int argc, char **argv)
{
  testing::InitGoogleTest(&argc, argv);

  return RUN_ALL_TESTS();
}
