//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------

#include "bodydynamics/factor/factor_force_torque_inertial_dynamics.h"
#include "bodydynamics/processor/processor_force_torque_inertial_preint_dynamics.h"
#include "bodydynamics/sensor/sensor_force_torque_inertial.h"
#include "bodydynamics/internal/config.h"

#include <core/sensor/factory_sensor.h>

#include <core/utils/utils_gtest.h>
#include <core/utils/logging.h>
#include <core/utils/params_server.h>
#include <core/yaml/parser_yaml.h>
#include <core/state_block/state_block_derived.h>
#include <core/state_block/factory_state_block.h>
#include <core/ceres_wrapper/solver_ceres.h>

#include <Eigen/Dense>
#include <Eigen/Geometry>

using namespace wolf;
using namespace bodydynamics::fti;

WOLF_REGISTER_STATEBLOCK_WITH_KEY(L, StateVector3d);

class Test_SolveProblemForceTorqueInertialPreintDynamics_yaml : public testing::Test
{
  public:
    ProblemPtr                                    P;
    SensorForceTorqueInertialPtr                  S;
    ProcessorForceTorqueInertialPreintDynamicsPtr p;
    SolverCeresPtr                                solver;
    Vector3d                                      cdm_true;
    Vector3d                                      inertia_true;
    double                                        mass_true;

  protected:
    void SetUp() override
    {
        std::string wolf_root = _WOLF_BODYDYNAMICS_ROOT_DIR;
        ParserYaml  parser =
            ParserYaml("problem_force_torque_inertial_dynamics_solve_test.yaml", wolf_root + "/test/yaml/");
        ParamsServer server = ParamsServer(parser.getParams());
        P                   = Problem::autoSetup(server);
        // CERES WRAPPER
        solver = std::make_shared<SolverCeres>(P);
        // solver->getSolverOptions().max_num_iterations               = 1e4;
        // solver->getSolverOptions().function_tolerance               = 1e-15;
        // solver->getSolverOptions().gradient_tolerance               = 1e-15;
        // solver->getSolverOptions().parameter_tolerance              = 1e-15;

        S = std::static_pointer_cast<SensorForceTorqueInertial>(P->getHardware()->getSensorList().front());
        p = std::static_pointer_cast<ProcessorForceTorqueInertialPreintDynamics>(S->getProcessorList().front());

        cdm_true     = {0, 0, 0.0341};
        inertia_true = {0.017598, 0.017957, 0.029599};
        mass_true    = 1.952;
    }
};

TEST_F(Test_SolveProblemForceTorqueInertialPreintDynamics_yaml, force_registraion)
{
    FactorySensor::registerCreator("SensorForceTorqueInertial", SensorForceTorqueInertial::create);
}

// Hover test.
TEST_F(Test_SolveProblemForceTorqueInertialPreintDynamics_yaml, mass_only_hovering)
{
    double mass_guess = S->getStateBlock('m')->getState()(0);

    // Data
    VectorXd data      = VectorXd::Zero(12);      // [ a, w, f, t ]
    data.segment<3>(0) = -gravity();              // accelerometer
    data.segment<3>(6) = -mass_true * gravity();  // force
    MatrixXd data_cov  = MatrixXd::Identity(12, 12) * 1e-6;

    // We fix the parameters of the sensor except for the mass
    S->getStateBlock('P')->fix();
    S->getStateBlock('O')->fix();
    S->getStateBlock('I')->fix();
    S->getStateBlock('C')->fix();
    S->getStateBlock('i')->fix();
    S->getStateBlock('m')->unfix();
    // S->setStateBlockDynamic('I');

    CaptureMotionPtr C0_0 = std::make_shared<CaptureMotion>("CaptureMotion", 0.0, S, data, data_cov, nullptr);
    CaptureMotionPtr C1_0 = std::make_shared<CaptureMotion>("CaptureMotion", 0.2, S, data, data_cov, nullptr);
    CaptureMotionPtr C2_0 = std::make_shared<CaptureMotion>("CaptureMotion", 0.4, S, data, data_cov, nullptr);
    CaptureMotionPtr C3_0 = std::make_shared<CaptureMotion>("CaptureMotion", 0.6, S, data, data_cov, nullptr);
    CaptureMotionPtr C4_0 = std::make_shared<CaptureMotion>("CaptureMotion", 0.8, S, data, data_cov, nullptr);
    CaptureMotionPtr C5_1 = std::make_shared<CaptureMotion>("CaptureMotion", 1.0, S, data, data_cov, nullptr);

    C0_0->process();
    CaptureMotionPtr C_KF0 = std::static_pointer_cast<CaptureMotion>(p->getOrigin());
    C1_0->process();
    C2_0->process();
    C3_0->process();
    C4_0->process();
    C5_1->process();

    CaptureMotionPtr C_KF1 = std::static_pointer_cast<CaptureMotion>(p->getOrigin());

    C_KF0->getFrame()->fix();
    C_KF1->getFrame()->fix();

    P->print(4, 1, 1, 1);

    WOLF_INFO("======== SOLVE PROBLEM =======")
    std::string report = solver->solve(wolf::SolverManager::ReportVerbosity::FULL);
    WOLF_INFO(report);

    WOLF_INFO("True mass     : ", mass_true, " Kg.");
    WOLF_INFO("Guess mass    : ", mass_guess, " Kg.");
    WOLF_INFO("Estimated mass: ", S->getStateBlock('m')->getState(), " Kg.");
    WOLF_INFO("-----------------------------");


    CaptureMotionPtr C = std::make_shared<CaptureMotion>("CaptureMotion", 2.0, S, data, data_cov, nullptr);
    for (TimeStamp t = p->getLast()->getTimeStamp() + 1.0 ; t <= 7.0 ; t += 1.0)
    {
        C->setTimeStamp(t);
        C->process();
        p->getOrigin()->getFrame()->fix();
        report = solver->solve(wolf::SolverManager::ReportVerbosity::FULL);
        WOLF_INFO("Estimated mass: ", S->getStateBlock('m')->getState(), " Kg.");
        WOLF_INFO("-----------------------------");
    }

    P->print(4, 1, 1, 1);


    ASSERT_NEAR(S->getStateBlock('m')->getState()(0), mass_true, 1e-3);
}

TEST_F(Test_SolveProblemForceTorqueInertialPreintDynamics_yaml, cdm_only_hovering)
{
    S->getStateBlock('C')->setState(Vector3d(0.01, 0.01, 0.033));
    S->getStateBlock('m')->setState(Vector1d(mass_true));
    Vector3d cdm_guess = S->getStateBlock('C')->getState();

    // Data
    VectorXd data      = VectorXd::Zero(12);  // [ a, w, f, t ]
    data.segment<3>(0) = -gravity();
    data.segment<3>(6) = -mass_true * gravity();
    MatrixXd data_cov  = MatrixXd::Identity(12, 12);

    // We fix the parameters of the sensor except for the cdm
    S->getStateBlock('P')->fix();
    S->getStateBlock('O')->fix();
    S->getStateBlock('I')->fix();
    S->getStateBlock('C')->unfix();
    S->getStateBlock('i')->fix();
    S->getStateBlock('m')->fix();

    CaptureMotionPtr C0_0 = std::make_shared<CaptureMotion>("CaptureMotion", 0.0, S, data, data_cov, nullptr);
    CaptureMotionPtr C1_0 = std::make_shared<CaptureMotion>("CaptureMotion", 0.2, S, data, data_cov, nullptr);
    CaptureMotionPtr C2_0 = std::make_shared<CaptureMotion>("CaptureMotion", 0.4, S, data, data_cov, nullptr);
    CaptureMotionPtr C3_0 = std::make_shared<CaptureMotion>("CaptureMotion", 0.6, S, data, data_cov, nullptr);
    CaptureMotionPtr C4_0 = std::make_shared<CaptureMotion>("CaptureMotion", 0.8, S, data, data_cov, nullptr);
    CaptureMotionPtr C5_1 = std::make_shared<CaptureMotion>("CaptureMotion", 1.0, S, data, data_cov, nullptr);

    C0_0->process();
    CaptureMotionPtr C_KF0 = std::static_pointer_cast<CaptureMotion>(p->getOrigin());
    C1_0->process();
    C2_0->process();
    C3_0->process();
    C4_0->process();
    C5_1->process();

    CaptureMotionPtr C_KF1 = std::static_pointer_cast<CaptureMotion>(p->getOrigin());

    C_KF0->getFrame()->fix();
    C_KF1->getFrame()->getStateBlock('L')->setState(Vector3d::Zero()); // Make sure L is not moving due to torques
    C_KF1->getFrame()->fix();

    P->print(4, 1, 1, 1);

    WOLF_INFO("======== SOLVE PROBLEM =======")
    std::string report = solver->solve(wolf::SolverManager::ReportVerbosity::FULL);
    WOLF_INFO(report);  // should show a very low iteration number (possibly 1)

    WOLF_INFO("True cdm     : ", cdm_true.transpose(), " m.");
    WOLF_INFO("Guess cdm    : ", cdm_guess.transpose(), " m.");
    WOLF_INFO("Estimated cdm: ", S->getStateBlock('C')->getState().transpose(), " m.");
    WOLF_INFO("-----------------------------");

    CaptureMotionPtr C = std::make_shared<CaptureMotion>("CaptureMotion", 2.0, S, data, data_cov, nullptr);
    for (TimeStamp t = p->getLast()->getTimeStamp() + 1.0 ; t <= 11.0 ; t += 1.0)
    {
        C->setTimeStamp(t);
        C->process();
        p->getOrigin()->getFrame()->getStateBlock('L')->setState(Vector3d::Zero()); // Make sure L is not moving due to torques
        p->getOrigin()->getFrame()->fix();
        report = solver->solve(wolf::SolverManager::ReportVerbosity::FULL);
        WOLF_INFO("Estimated cdm : ", S->getStateBlock('C')->getState().transpose(), " m.");
        WOLF_INFO("-----------------------------");
    }


    ASSERT_MATRIX_APPROX(S->getStateBlock('C')->getState().head(2), cdm_true.head(2), 1e-5); // cdm_z is not observable while hovering
}

TEST_F(Test_SolveProblemForceTorqueInertialPreintDynamics_yaml, mass_and_cdm_hovering)
{
    S->getStateBlock('C')->setState(Vector3d(0.005, 0.005, 0.05));
    S->getStateBlock('m')->setState(Vector1d(1.900));
    Vector3d cdm_guess  = S->getStateBlock('C')->getState();
    double   mass_guess = S->getStateBlock('m')->getState()(0);

    // Data
    VectorXd data      = VectorXd::Zero(12);  // [ a, w, f, t ]
    data.segment<3>(0) = -gravity();
    data.segment<3>(6) = -mass_true * gravity();
    MatrixXd data_cov  = MatrixXd::Identity(12, 12);

    // We fix the parameters of the sensor except for the cdm and the mass
    S->getStateBlock('P')->fix();
    S->getStateBlock('O')->fix();
    S->getStateBlock('I')->fix();
    S->getStateBlock('C')->unfix();
    S->getStateBlock('i')->fix();
    S->getStateBlock('m')->unfix();

    CaptureMotionPtr C0_0 = std::make_shared<CaptureMotion>("CaptureMotion", 0.0, S, data, data_cov, nullptr);
    CaptureMotionPtr C1_0 = std::make_shared<CaptureMotion>("CaptureMotion", 0.2, S, data, data_cov, nullptr);
    CaptureMotionPtr C2_0 = std::make_shared<CaptureMotion>("CaptureMotion", 0.4, S, data, data_cov, nullptr);
    CaptureMotionPtr C3_0 = std::make_shared<CaptureMotion>("CaptureMotion", 0.6, S, data, data_cov, nullptr);
    CaptureMotionPtr C4_0 = std::make_shared<CaptureMotion>("CaptureMotion", 0.8, S, data, data_cov, nullptr);
    CaptureMotionPtr C5_1 = std::make_shared<CaptureMotion>("CaptureMotion", 1.0, S, data, data_cov, nullptr);

    C0_0->process();
    CaptureMotionPtr C_KF0 = std::static_pointer_cast<CaptureMotion>(p->getOrigin());
    C1_0->process();
    C2_0->process();
    C3_0->process();
    C4_0->process();
    C5_1->process();

    CaptureMotionPtr C_KF1 = std::static_pointer_cast<CaptureMotion>(p->getOrigin());

    C_KF0->getFrame()->fix();
    C_KF1->getFrame()->getStateBlock('L')->setState(Vector3d::Zero()); // Make sure L is not moving due to torques
    C_KF1->getFrame()->fix();

    P->print(4, 1, 1, 1);

    WOLF_INFO("======== SOLVE PROBLEM =======")
    std::string report = solver->solve(wolf::SolverManager::ReportVerbosity::FULL);
    WOLF_INFO(report);  // should show a very low iteration number (possibly 1)

    WOLF_INFO("True mass     : ", mass_true, " Kg.");
    WOLF_INFO("True cdm      : ", cdm_true.transpose(), " m.");
    WOLF_INFO("Guess mass    : ", mass_guess, " Kg.");
    WOLF_INFO("Guess cdm     : ", cdm_guess.transpose(), " m.");
    WOLF_INFO("Estimated mass: ", S->getStateBlock('m')->getState(), " Kg.");
    WOLF_INFO("Estimated cdm : ", S->getStateBlock('C')->getState().transpose(), " m.");
    WOLF_INFO("-----------------------------");


    // Do a number of iterations with more keyframes, solving so that the calibration gets better and better
    // Here we take advantage of the sliding window, thereby getting rid progressively 
    // of the old factors, which contained calibration parameters far from the converged values,
    // which if not eliminated contaminate the overall solution.
    // This is due to these first factors relying on the linearization Jacobian to correct the
    // poorly preintegrated delta.
    CaptureMotionPtr C = std::make_shared<CaptureMotion>("CaptureMotion", 2.0, S, data, data_cov, nullptr);
    for (TimeStamp t = p->getLast()->getTimeStamp() + 1.0 ; t <= 7.0 ; t += 1.0)
    {
        C->setTimeStamp(t);
        C->process();
        p->getOrigin()->getFrame()->getStateBlock('L')->setState(Vector3d::Zero()); // Make sure L is not moving due to torques
        p->getOrigin()->getFrame()->fix();
        report = solver->solve(wolf::SolverManager::ReportVerbosity::FULL);
        WOLF_INFO("Estimated mass: ", S->getStateBlock('m')->getState(), " Kg.");
        WOLF_INFO("Estimated cdm : ", S->getStateBlock('C')->getState().transpose(), " m.");
        WOLF_INFO("-----------------------------");
    }

    auto cdm_estimated  = S->getStateBlock('C')->getState();
    auto mass_estimated = S->getStateBlock('m')->getState()(0);


    ASSERT_NEAR(mass_estimated, mass_true, 1e-6);
    ASSERT_MATRIX_APPROX(cdm_estimated.head(2), cdm_true.head(2), 1e-6); // cdm_z is not observable while hovering
}

int main(int argc, char **argv)
{
    testing::InitGoogleTest(&argc, argv);
    // ::testing::GTEST_FLAG(filter) = "Test_SolveProblemForceTorqueInertialPreintDynamics_yaml.cdm_only_hovering";
    return RUN_ALL_TESTS();
}
